open B8_types
open B8_utils

let not_found = Not_found
  
let rec iter_merge tdoc last =
  match tdoc.expr with
    Ttag_B tdoc2 -> let (last, tdoc2) = iter_merge tdoc2 last in
      let tdoc = { tdoc with expr = Ttag_B tdoc2 } in
      last, tdoc
  | Ttag_EM tdoc2 -> let (last, tdoc2) = iter_merge tdoc2 last in
      let tdoc = { tdoc with expr = Ttag_EM tdoc2 } in
      last, tdoc
  | Ttag_I tdoc2 -> let (last, tdoc2) = iter_merge tdoc2 last in
      let tdoc = { tdoc with expr = Ttag_I tdoc2 } in
      last, tdoc
  | Ttag_PL tdoc2 -> let (last, tdoc2) = iter_merge tdoc2 last in
      let tdoc = { tdoc with expr = Ttag_PL tdoc2 } in
      last, tdoc
  | Ttag_S tdoc2 -> let (last, tdoc2) = iter_merge tdoc2 last in
      let tdoc = { tdoc with expr = Ttag_S tdoc2 } in
      last, tdoc
  | Ttag_TT tdoc2 -> let (last, tdoc2) = iter_merge tdoc2 last in
      let tdoc = { tdoc with expr = Ttag_TT tdoc2 } in
      last, tdoc
  | Ttag_U tdoc2 -> let (last, tdoc2) = iter_merge tdoc2 last in
      let tdoc = { tdoc with expr = Ttag_U tdoc2 } in
      last, tdoc
  | Ttag_color (tdoc2, color)  -> let (last, tdoc2) = iter_merge tdoc2 last in
      let tdoc = { tdoc with expr = Ttag_color (tdoc2, color) } in
      last, tdoc
  | Ttag_size (tdoc2, size) -> let (last, tdoc2) = iter_merge tdoc2 last in
      let tdoc = { tdoc with expr = Ttag_size (tdoc2, size) } in
      last, tdoc
  | Tlist list -> 
      let last, list = List.fold_left (fun (last, list) tdoc ->
            try
              let last, tdoc = iter_merge tdoc last in
              (last, tdoc :: list)
            with _ ->
                last, list
        ) (last, []) list in
      if list = [] then raise not_found else
        last, { tdoc with expr = Tlist list }
  | Tstring "" -> raise not_found
  | Tstring s -> 
      match last with
        None -> let tdoc = { tdoc with expr = Tstring s } in
          (Some tdoc), tdoc
      | Some ({ expr = Tstring s2 } as tdoc2) ->
          if tdoc.prop = tdoc2.prop then
            (tdoc2.expr <- Tstring (s2 ^ s); raise not_found)
          else
          let tdoc = { tdoc with expr = Tstring s } in
          (Some tdoc), tdoc
      | _ -> assert false

let merge tdoc =
  try
    let _, tdoc = iter_merge tdoc None in
    tdoc
  with _ -> { expr = Tlist [] ; prop = prop_root }
  
type flat = {
    strs : string array;
    ints : int array;
    len : int;
    pos : int;
  }
  
  
let flatten_tdoc tdoc =
  let str_tab = ref [||] in
  let int_tab = ref [||] in
  let tab_len = ref 0 in
  let tab_pos = ref 0 in
  
  let rec iter tdoc =
    match tdoc.expr with
      Ttag_B tdoc
    | Ttag_EM tdoc
    | Ttag_I tdoc
    | Ttag_PL tdoc
    | Ttag_S tdoc
    | Ttag_TT tdoc
    | Ttag_U tdoc
    | Ttag_color (tdoc, _)
    | Ttag_size (tdoc, _) -> iter tdoc
    | Tlist list -> List.iter iter list
    | Tstring "" -> ()
    | Tstring string -> 
        let prop = int_of_prop tdoc.prop in
(*        Printf.printf "pos: %d len:%d" !tab_pos !tab_len; print_newline ();*)
        if !tab_pos = !tab_len then
          let new_len = !tab_len * 2 + 50 in
          let new_str_tab = Parray.create new_len string in
          let new_int_tab = Iarray.create new_len prop in
          Parray.blit !str_tab 0 new_str_tab 0 !tab_pos;
          Iarray.blit !int_tab 0 new_int_tab 0 !tab_pos;
          int_tab := new_int_tab;
          str_tab := new_str_tab;
          tab_len := new_len;
          incr tab_pos
        else
        let pos = !tab_pos in
        (!str_tab).(pos) <- string;
        (!int_tab).(pos) <- prop;
        tab_pos := pos + 1
  in
  iter tdoc;
  { strs = !str_tab;
    ints = !int_tab;
    len = !tab_pos;
    pos = 0;}

let merges = ref 0
  
let rec merge_strs t pos1 pos2 =
  if t.len = pos1 then
    { t with len = pos2 }
  else
  if t.ints.(pos1) = t.ints.(pos2) then begin
      t.strs.(pos2) <- t.strs.(pos2) ^ t.strs.(pos1);
      incr merges;
      merge_strs t (pos1+1) pos2
    end
  else begin
      let pos2 = pos2+1 in
      if pos1 <> pos2 then begin
          t.strs.(pos2) <- t.strs.(pos1);
          t.ints.(pos2) <- t.ints.(pos1);
        end;
      merge_strs t (pos1+1) pos2
    end
    
let flat_and_build tdoc =
  let t = flatten_tdoc tdoc in
  let t = merge_strs t 1 0 in
(*  Printf.printf "Merges %d" !merges; print_newline ();*)
  tdoc

  (*
let _ = 
  register "b8_flatten" flat_and_build 10;
  register "merge" merge 5
*)