open B8_types
  
let char_of_color color =
  match color with
    Red -> 'r'
  | Green -> 'g'
  | Blue -> 'b'
  | Cyan -> 'c'
  | Magenta -> 'm'
  | Yellow -> 'y'
  | Black -> 'k'
  | White -> 'w'
  | Unspecified -> 'u'

let char_of_size size =
  match size with
    T0 -> '0'
  | T1 -> '1'
  | T2 -> '2'
  | T3 -> '3'
  | T4 -> '4'
  | T5 -> '5'
  | T6 -> '6'
  | T7 -> '7'
  | T8 -> '8'
  | T9 -> '9'
  | Tnone -> '-'

let rec dump_doc tdoc =
  match tdoc with
    Dtag_B tdoc -> 
      print_string "<B>"; dump_doc tdoc; print_string "</B>"
  | Dtag_EM tdoc -> 
      print_string "<EM>"; dump_doc tdoc; print_string "</EM>"
  | Dtag_I tdoc -> 
      print_string "<I>"; dump_doc tdoc; print_string "</I>"
  | Dtag_PL tdoc -> 
      print_string "<PL>"; dump_doc tdoc; print_string "</PL>"
  | Dtag_S tdoc -> 
      print_string "<S>"; dump_doc tdoc; print_string "</S>"
  | Dtag_TT tdoc -> 
      print_string "<TT>"; dump_doc tdoc; print_string "</TT>"
  | Dtag_U tdoc -> 
      print_string "<U>"; dump_doc tdoc; print_string "</U>"
  | Dtag_color (tdoc, color) -> 
      let color = char_of_color color in
      print_char '<';
      print_char color;
      print_char '>';
      dump_doc tdoc; 
      print_char '<';
      print_char '/';
      print_char color;
      print_char '>';
  | Dtag_size (tdoc, size) -> 
      let size = char_of_size size in
      print_char '<';
      print_char size;
      print_char '>';
      dump_doc tdoc; 
      print_char '<';
      print_char '/';
      print_char size;
      print_char '>';
  | Dlist list ->
      dump_docs list    
  | Dstring string -> print_string string

and dump_docs list =
  match list with
    [] -> ()
  | tdoc :: tail ->
      dump_doc tdoc;
      dump_docs tail

      
let dump_prop prop =
  print_string "{ ";
  if prop.b then print_string "B ";
  if prop.em then print_string "EM ";
  if prop.i then print_string "I ";
  if prop.s then print_string "S ";
  if prop.tt then print_string "TT ";
  for i = 0 to prop.u - 1 do print_char 'U'; done;
  if prop.u > 0 then print_char ' ';
  if prop.size <> Tnone then (print_char (char_of_size prop.size); print_char ' ');
  if prop.color <> Unspecified then 
    (print_char (char_of_color prop.color); print_char ' ');
  print_char '}'

let rec dump_tdoc tdoc =
  print_newline ();
  dump_prop tdoc.prop;
  print_newline ();
  (match tdoc.expr with
      Ttag_B tdoc -> 
        print_string "<B>"; dump_tdoc tdoc; print_string "</B>"
    | Ttag_EM tdoc -> 
        print_string "<EM>"; dump_tdoc tdoc; print_string "</EM>"
    | Ttag_I tdoc -> 
        print_string "<I>"; dump_tdoc tdoc; print_string "</I>"
    | Ttag_PL tdoc -> 
        print_string "<PL>"; dump_tdoc tdoc; print_string "</PL>"
    | Ttag_S tdoc -> 
        print_string "<S>"; dump_tdoc tdoc; print_string "</S>"
    | Ttag_TT tdoc -> 
        print_string "<TT>"; dump_tdoc tdoc; print_string "</TT>"
    | Ttag_U tdoc -> 
        print_string "<U>"; dump_tdoc tdoc; print_string "</U>"
    | Ttag_color (tdoc, color) -> 
        let color = char_of_color color in
        print_char '<';
        print_char color;
        print_char '>';
        dump_tdoc tdoc; 
        print_char '<';
        print_char '/';
        print_char color;
        print_char '>';
    | Ttag_size (tdoc, size) -> 
        let size = char_of_size size in
        print_char '<';
        print_char size;
        print_char '>';
        dump_tdoc tdoc; 
        print_char '<';
        print_char '/';
        print_char size;
        print_char '>';
    | Tlist list ->
        Printf.printf "<LIST>";
        dump_tdocs list;
        Printf.printf "</LIST>"
    | Tstring string -> 
        print_string "<STRING>"; print_string string; 
        print_string "</STRING>")

and dump_tdocs list =
  match list with
    [] -> ()
  | tdoc :: tail ->
      dump_tdoc tdoc;
      dump_tdocs tail

let rec print_tdoc tdoc =
  match tdoc.expr with
      Ttag_B tdoc -> 
        print_string "<B>"; print_tdoc tdoc; print_string "</B>"
    | Ttag_EM tdoc -> 
        print_string "<EM>"; print_tdoc tdoc; print_string "</EM>"
    | Ttag_I tdoc -> 
        print_string "<I>"; print_tdoc tdoc; print_string "</I>"
    | Ttag_PL tdoc -> 
        print_string "<PL>"; print_tdoc tdoc; print_string "</PL>"
    | Ttag_S tdoc -> 
        print_string "<S>"; print_tdoc tdoc; print_string "</S>"
    | Ttag_TT tdoc -> 
        print_string "<TT>"; print_tdoc tdoc; print_string "</TT>"
    | Ttag_U tdoc -> 
        print_string "<U>"; print_tdoc tdoc; print_string "</U>"
    | Ttag_color (tdoc, color) -> 
        let color = char_of_color color in
        print_char '<';
        print_char color;
        print_char '>';
        print_tdoc tdoc; 
        print_char '<';
        print_char '/';
        print_char color;
        print_char '>';
    | Ttag_size (tdoc, size) -> 
        let size = char_of_size size in
        print_char '<';
        print_char size;
        print_char '>';
        print_tdoc tdoc; 
        print_char '<';
        print_char '/';
        print_char size;
        print_char '>';
    | Tlist list ->
        Printf.printf "<LIST>";
        print_tdocs list;
        Printf.printf "</LIST>"
    | Tstring string -> 
        print_string string
  
      
and print_tdocs list =
  match list with
    [] -> ()
  | tdoc :: tail ->
      print_tdoc tdoc;
      print_tdocs tail
