(* size nesting *)

open B8_types

exception Sizenestingmerge
exception Sizenestingmergedeep

let rec flatlists_tdoc t =
  match t.expr with
    Ttag_color(d,c) -> { expr = Ttag_color(flatlists_tdoc d , c ); prop = t.prop }
  | Ttag_B(d) -> { expr = Ttag_B(flatlists_tdoc d) ; prop = t.prop }
  | Ttag_EM(d) -> {expr = Ttag_EM(flatlists_tdoc d) ; prop = t.prop }
  | Ttag_I(d) -> {expr = Ttag_I(flatlists_tdoc d) ; prop = t.prop }
  | Ttag_PL(d) -> {expr = Ttag_PL(flatlists_tdoc d) ; prop = t.prop }
  | Ttag_S(d) -> {expr = Ttag_S(flatlists_tdoc d); prop = t.prop }
  | Ttag_TT(d) -> {expr = Ttag_TT(flatlists_tdoc d) ; prop = t.prop }
  | Ttag_U(d) -> {expr = Ttag_U(flatlists_tdoc d) ; prop = t.prop }
  | Ttag_size(d,s) -> { expr = Ttag_size((flatlists_tdoc d) , s) ; prop = t.prop }
  | Tlist(dl) -> 
      let f = List.map flatlists_tdoc dl in
      let cat x y = 
	begin
	  match x.expr with
	    Tlist(tdl)-> tdl@y
	  | _ -> x::y 
	end in
      {expr = Tlist(List.fold_right cat f []) ; prop = t.prop }
  | _ -> t
    

let rec merge_tdoc td1 td2 p=
  match td1,td2 with
    Ttag_size(d1,c1) , Ttag_size(d2,c2) when c1=c2 -> 
      let td = flatlists_tdoc {expr = Tlist([d1;d2]) ; prop = {p with size = c1} } in
      Ttag_size(td,c1)
  | _ -> raise Sizenestingmerge


let rec size_nesting td = 
  match td.expr with
    Ttag_size(d,c) -> {expr=Ttag_size(size_nesting d,c) ; prop = td.prop } 
  | Tlist(dl) -> {expr = Tlist(wrapper_adjacent_size (List.map size_nesting dl)) ; prop = td.prop } 
  | Ttag_B(d) -> 
      (let td' = size_nesting d in 
      match  td'.expr with
	Ttag_size(d',c') -> 
	  let x = {expr = Ttag_B(d') ; prop = {td.prop with size = c'} } in
	  {expr = Ttag_size( x ,c') ; prop = td.prop  } 
      |	_ -> {expr = Ttag_B(td') ; prop = td.prop } )

  | Ttag_EM(d) -> 
      (let td' = size_nesting d in 
      match  td'.expr with
	Ttag_size(d',c') -> 
	  let x = {expr = Ttag_EM(d') ; prop = {td.prop with size = c'} } in
	  {expr = Ttag_size( x ,c') ; prop = td.prop  } 
      |	_ -> {expr = Ttag_EM(td') ; prop = td.prop } )

  | Ttag_I(d) -> 
      (let td' = size_nesting d in 
      match  td'.expr with
	Ttag_size(d',c') -> 
	  let x = {expr = Ttag_I(d') ; prop = {td.prop with size = c'} } in
	  {expr = Ttag_size( x ,c') ; prop = td.prop  } 
      |	_ -> {expr = Ttag_I(td') ; prop = td.prop } )

  | Ttag_PL(d) -> 
      (let td' = size_nesting d in 
      match  td'.expr with
	Ttag_size(d',c') -> 
	  let x = {expr = Ttag_PL(d') ; prop = {td.prop with size = c'} } in
	  {expr = Ttag_size( x ,c') ; prop = td.prop  } 
      |	_ -> {expr = Ttag_PL(td') ; prop = td.prop } )

  | Ttag_S(d) -> 
      (let td' = size_nesting d in 
      match  td'.expr with
	Ttag_size(d',c') -> 
	  let x = {expr = Ttag_S(d') ; prop = {td.prop with size = c'} } in
	  {expr = Ttag_size( x ,c') ; prop = td.prop  } 
      |	_ -> {expr = Ttag_S(td') ; prop = td.prop } )

  | Ttag_TT(d) -> 
      (let td' = size_nesting d in 
      match  td'.expr with
	Ttag_size(d',c') -> 
	  let x = {expr = Ttag_TT(d') ; prop = {td.prop with size = c'} } in
	  {expr = Ttag_size( x ,c') ; prop = td.prop  } 
      |	_ -> {expr = Ttag_TT(td') ; prop = td.prop } )

  | Ttag_U(d) -> 
      (let td' = size_nesting d in 
      match  td'.expr with
	Ttag_size(d',c') -> 
	  let x = {expr = Ttag_U(d') ; prop = {td.prop with size = c'} } in
	  {expr = Ttag_size( x ,c') ; prop = td.prop  } 
      |	_ -> {expr = Ttag_U(td') ; prop = td.prop } )

  | Ttag_color(d,s) -> 
      (let td' = size_nesting d in 
      match  td'.expr with
	Ttag_size(d',c') -> 
	  let x = {expr = Ttag_color(d',s) ; prop = {td.prop with size = c'} } in
	  {expr = Ttag_size( x ,c') ; prop = td.prop  } 
      |	_ -> {expr = Ttag_color(td',s) ; prop = td.prop } )

  | _ -> td

and wrapper_adjacent_size tdl = 
  match tdl with
    [] -> []
  | x::l -> 
      match x.expr with
	Ttag_size(d,c) -> adjacent_size x l
      |	_ -> x::(wrapper_adjacent_size l)

and merge_deep tde1 tde2 c p =
(*  print_string "tde1:"; B8_dump.dump_tdoc {expr=tde1;prop=p} ; print_newline(); 
  print_string "tde2:"; B8_dump.dump_tdoc {expr=tde2;prop=p} ; print_newline();*)
  match tde1 with
    Ttag_size(d,c') -> 
      if c=c' then  merge_tdoc tde1 tde2 p
      else
	begin
	  let tde2expr = {expr = tde2 ; prop = {p with size = c'}} in
	  match d.expr with
	    Tlist(dl) -> 
	      begin
		let rec trav  = function
		    [] -> [tde2expr]
		  |	[{expr=Ttag_size(d3,c3) as tde3; prop=p3}] -> [{expr= (merge_deep tde3 tde2 c p) ; prop = p }]
		  |	x::l -> x::(trav l)
		in Ttag_size({expr = Tlist(trav dl) ; prop = p} , c')
	      end
	  | _ -> 
	      let dexpr = {expr = d.expr ; prop = {p with size = c'} } in
	      Ttag_size({expr = Tlist([dexpr;tde2expr]) ; prop = p } ,c')
	end
  | _ ->  raise Sizenestingmergedeep


and adjacent_size x tdl = 
  match tdl with
    [] -> [x]
  | t::l -> 
      match t.expr with
	Ttag_size(d,c') -> 
	  adjacent_size {expr=(merge_deep x.expr t.expr c' x.prop) ; prop = x.prop}  l
      |	_ -> x::t::(wrapper_adjacent_size l)



let no_opt td = td

let _ = B8_types.register "size_nesting" size_nesting 0  

let _ = B8_types.register "no_opt" no_opt 0 

