open B8_types

(* supprime les espaces clairement inutiles *)
let delete_white s =
  let len = (String.length s) in
  let res = Buffer.create len in
  let rec aux i spc =
    if i < len
    then
      match s.[i] with
      | ' ' | '\r' | '\n' | '\t' ->
          if spc
          then (* forget multiple spaces *)
            aux (i+1) true
          else
            (Buffer.add_char res ' ';
             aux (i+1) true)
      | x ->
          Buffer.add_char res x; aux (i+1) false in
  aux 0 false;
  Buffer.contents res

let convert_spaces s =
  for i = 0 to String.length s - 1 do
    match s.[i] with
      '\r' -> s.[i] <- ' '
    |  _ -> ()
  done
  
(* rajoute des etiquettes a chaque noeud d'un arbre *)
let rec transf prop = function
  | Dtag_B d ->
      if prop.b
      then transf prop d
      else
        (let np = { prop with b = true } in
        { expr = Ttag_B (transf np d); prop = prop })
  | Dtag_EM d ->
      if prop.s
      then transf prop d
      else
        (let np = { prop with em = not prop.em } in
        { expr = Ttag_EM (transf np d); prop = prop })
  | Dtag_I d ->
      if prop.i
      then transf prop d
      else
        (let np = { prop with i = true } in
        { expr = Ttag_I (transf np d); prop = prop })
  | Dtag_PL d ->
      if prop.u != 0 || prop.s || prop.em || prop.i || prop.b || prop.tt
      then
        (let np = { prop_root with size = prop.size ; color = prop.color } in
        { expr = Ttag_PL (transf np d); prop = prop })
      else transf prop d
  | Dtag_S d ->
      if prop.s
      then transf prop d
      else
        (let np = { prop with s = true; em = false } in
        { expr = Ttag_S (transf np d); prop = prop })
  | Dtag_TT d ->
      if prop.tt
      then transf prop d
      else
        (let np = { prop with tt = true } in
        { expr = Ttag_TT (transf np d); prop = prop })
  | Dtag_U d ->
      if prop.u == 3
      then transf prop d
      else
        (let np = { prop with u = prop.u+1 } in
        { expr = Ttag_U (transf np d); prop = prop })
  | Dtag_color (d, c) ->
      if prop.color == c
      then transf prop d
      else
        (let np = { prop with color = c } in
        { expr = Ttag_color (transf np d, c); prop = prop })
  | Dtag_size (d, t) ->
      if prop.size == t
      then transf prop d
      else
        (let np = { prop with size = t } in
        { expr = Ttag_size (transf np d, t); prop = prop })
  | Dlist l -> { expr = Tlist (List.map (transf prop) l); prop = prop }
  | Dstring s ->
      if prop.tt
      then { expr = Tstring s; prop = prop }
      else { expr = Tstring (delete_white s); prop = prop }

let typedoc d = transf prop_root d
