open B8_types
  
(* memoization des properties *)  
let int_of_color color =
  match color with
    Red -> 0
  | Green -> 1
  | Blue -> 2
  | Cyan -> 3
  | Magenta -> 4
  | Yellow -> 5
  | Black -> 6
  | White -> 7
  | Unspecified -> 8

let int_of_size size =
  match size with
    T0 -> 0
  | T1 -> 1
  | T2 -> 2
  | T3 -> 3
  | T4 -> 4
  | T5 -> 5
  | T6 -> 6
  | T7 -> 7
  | T8 -> 8
  | T9 -> 9
  | Tnone -> 10
      
let color_of_int color =
  match color with
    0 -> Red
  | 1 -> Green
  | 2 -> Blue
  | 3 -> Cyan
  | 4 -> Magenta
  | 5 -> Yellow
  | 6 -> Black
  | 7 -> White
  | 8 -> Unspecified
  | _ -> assert false
      
let size_of_int size =
  match size with
    0 -> T0
  | 1 -> T1
  | 2 -> T2
  | 3 -> T3
  | 4 -> T4
  | 5 -> T5
  | 6 -> T6
  | 7 -> T7
  | 8 -> T8
  | 9 -> T9
  | 10 -> Tnone
  | _ -> assert false
      
let map = ref Imap.empty
  
let int_of_prop prop =
  let int =
    (if prop.b then 1 else 0) +
      (if prop.em then 2 else 0) +
      (if prop.i then 4 else 0) +
      (if prop.s then 8 else 0) +
      (if prop.tt then 16 else 0) +
      (prop.u lsl 5) + (* 2 bits *)
      (int_of_size prop.size lsl 7) + (* 4 bits *)
      (int_of_color prop.color lsl 11) (* 4 bits *)
  in
  if not (Imap.mem int !map) then map := Imap.add int prop !map;
  int
  
let prop_of_int int = 
  {
    b = int land 1 <> 0;
    em = int land 2 <> 0;
    i = int land 4 <> 0;
    s = int land 8 <> 0;
    tt = int land 16 <> 0;
    u = (int lsr 5) land 3;
    size = size_of_int ((int lsr 7) land 15);
    color = color_of_int ((int lsr 11) land 15);
  }

let uniq_int_of_prop prop =
  let int = int_of_prop prop in
  if not (Imap.mem int !map) then map := Imap.add int prop !map;
  int
  
let uniq_prop_of_int int =  Imap.find int !map 
  
let rec flatten_tdoc tdoc =
  match tdoc.expr with
    Ttag_B tdoc
  | Ttag_EM tdoc
  | Ttag_I tdoc
  | Ttag_PL tdoc
  | Ttag_S tdoc
  | Ttag_TT tdoc
  | Ttag_U tdoc
  | Ttag_color (tdoc, _)
  | Ttag_size (tdoc, _) -> flatten_tdoc tdoc
  | Tlist list ->
      List.flatten (List.map flatten_tdoc list)
  | Tstring "" -> []
  | Tstring string -> [ { expr = tdoc.expr ; 
          prop = uniq_prop_of_int (int_of_prop tdoc.prop) } ]

let space_prop prop =
  { prop with
    s = false;
    em = false;
    i = false;
    b = false;
    color = if prop.u = 0 then White else prop.color;
  }
  
let u_int int = (int lsr 5) land 3
  
let n = 16 + (3 lsl 5) + (15 lsl 7)
let space_int int =
  int land (n + (if u_int int <> 0 then
        (15 lsl 11) else 0))

let is_tt int = int land 16 <> 0