open B8_prop
open String
open B8_types

open B8_dump
  open B8_utils
  
let validate _ _  = false
let rec size current tdoc =
  match tdoc.expr with
    Ttag_B tdoc 
  | Ttag_I tdoc
  | Ttag_S tdoc
  | Ttag_U tdoc
  | Ttag_color (tdoc,_) 
  | Ttag_size (tdoc,_) -> size (current + 7) tdoc
  | Ttag_EM tdoc
  | Ttag_PL tdoc
  | Ttag_TT tdoc ->  size (current + 9) tdoc
  | Tlist tdocs -> 
      List.fold_left size current tdocs
  | Tstring string -> current + String.length string
    
let size tdoc = size 0 tdoc

type ttdoc = {
    str : string;
    int : int
  }

type t = {
    tstr : string;
    tlen : int;
    tprop: int; (* prop are uniq *)
    tnext : flat;
  }

and flat = {
    strs : string array;
    ints : int array;
    len : int;
    mutable pos : int;
  }

let succ t = t.pos <- t.pos + 1
  
  
external check : bool -> bool = "%identity"
  

let flatten_tdoc tdoc =
  let str_tab = ref [||] in
  let int_tab = ref [||] in
  let tab_len = ref 0 in
  let tab_pos = ref 0 in
  
  let rec iter tdoc =
    match tdoc.expr with
      Ttag_B tdoc
    | Ttag_EM tdoc
    | Ttag_I tdoc
    | Ttag_PL tdoc
    | Ttag_S tdoc
    | Ttag_TT tdoc
    | Ttag_U tdoc
    | Ttag_color (tdoc, _)
    | Ttag_size (tdoc, _) -> iter tdoc
    | Tlist list -> List.iter iter list
    | Tstring "" -> ()
    | Tstring string -> 
        let prop = int_of_prop tdoc.prop in
(*        Printf.printf "pos: %d len:%d" !tab_pos !tab_len; print_newline ();*)
        if !tab_pos = !tab_len then
          let new_len = !tab_len * 2 + 50 in
          let new_str_tab = Parray.create new_len string in
          let new_int_tab = Iarray.create new_len prop in
          Parray.blit !str_tab 0 new_str_tab 0 !tab_pos;
          Iarray.blit !int_tab 0 new_int_tab 0 !tab_pos;
          int_tab := new_int_tab;
          str_tab := new_str_tab;
          tab_len := new_len;
          incr tab_pos
        else
        let pos = !tab_pos in
        (!str_tab).(pos) <- string;
        (!int_tab).(pos) <- prop;
        tab_pos := pos + 1
  in
  iter tdoc;
  { strs = !str_tab;
    ints = !int_tab;
    len = !tab_pos;
    pos = 0;}
  
let rec only_spaces t pos p =
  if t.tlen = pos then
    let f = t.tnext in
    f.pos = f.len ||
    let npos = f.pos in
    let s = f.strs.(npos) in
    let pp = f.ints.(npos) in
    succ f;
    space_int pp = p &&
    only_spaces 
      { tstr = s; tprop = space_int pp; tlen = String.length s; tnext = f } 0 p
  else
  let c = t.tstr.[pos] in
  c = ' ' && only_spaces t (pos+1) p
  
let rec valid t1 pos1 t2 pos2 = (* previous char is not a space *)
  (*
  let check v =
    if not v then
      begin
        Printf.printf "Different"; print_newline ();
        Printf.printf "[%s][%d]" (escaped t1.tstr) pos1;  print_newline ();
        Printf.printf "[%s][%d]" (escaped t2.tstr) pos2;  print_newline ();
        dump_prop t1.tprop; print_newline ();
        dump_prop t2.tprop; print_newline ();
      end;
    v
in
  *)
  if t1.tlen = pos1 then
    let f1 = t1.tnext in
    if f1.pos = f1.len then
      let f2 = t2.tnext in
      check (t2.tlen = pos2 && f2.len = f2.pos)
    else
    let npos = f1.pos in
    succ f1;
    let s1 = f1.strs.(npos) in
    let p1 = f1.ints.(npos) in
    valid { tstr =s1; tprop = p1; tlen = String.length s1;
      tnext = f1 } 0 t2 pos2
  else
  if t2.tlen = pos2 then
    valid t2 pos2 t1 pos1
  else
  let cc1 = t1.tstr.[pos1] in
  let cc2 = t2.tstr.[pos2] in
  if cc1 = ' ' && cc2 = ' ' then
    let p1 = space_int t1.tprop in
    let p2 = space_int t2.tprop in
    check (p1 = p2) && 
    if is_tt p1 then
      valid t1 (pos1+1) t2 (pos2+1)
    else
      valid_spc t1 (pos1+1) t2 (pos2+1) p1
  else
    check ((cc1 = cc2 && t1.tprop == t2.tprop) ||
      (is_space cc1 && is_space cc2 &&
        space_int t1.tprop = space_int t2.tprop))
    
    &&
    valid t1 (pos1+1) t2 (pos2+1)

and valid_spc t1 pos1 t2 pos2 p =
    (*
  let check v =
    if not v then
      begin
        Printf.printf "Different (spaces)"; print_newline ();
        Printf.printf "[%s][%d]" (escaped t1.tstr) pos1;  print_newline ();
        Printf.printf "[%s][%d]" (escaped t2.tstr) pos2;  print_newline ();
        dump_prop t1.tprop; print_newline ();
        dump_prop t2.tprop; print_newline ();
      end;
    v
  in
*)
  if t1.tlen = pos1 then
    let f1 = t1.tnext in
    let npos = f1.pos in
    if npos = f1.len then
      check (t2.tlen = pos2 || space_int t2.tprop = p) &&
      only_spaces t2 pos2 p
    else
    let s1 = f1.strs.(npos) in
    let p1 = f1.ints.(npos) in
    succ f1;
    valid_spc { tstr =s1; tprop = p1; tlen = String.length s1;
      tnext = f1 } 0 t2 pos2 p
  else
  if t2.tlen = pos2 then
    valid_spc t2 pos2 t1 pos1 p
  else
  let cc1 = t1.tstr.[pos1] in
  let cc2 = t2.tstr.[pos2] in
  if cc1 = ' ' then begin
      let p1 = space_int t1.tprop in
      if p1 = p then
(* the space in p1 is irrelevant *)
        valid_spc t1 (pos1+1) t2 pos2 p
      else (* the space in s1 is relevant *)
        check (cc2 = ' ') &&
        let p2 = space_int t2.tprop in
        if p2 = p then (* the space in s2 is irrelevant *)
          valid_spc t2 (pos2+1) t1 pos1 p
        else (* the space in s2 is relevant too *)
          check (p1 = p2) &&
          if is_tt p1 then
            valid t1 (pos1+1) t2 (pos2+1)
          else
            valid_spc t1 (pos1+1) t2 (pos2+1) p1
    end
    else
    if cc2 = ' ' then
      let p2 = space_int t2.tprop in
      if p2 = p then (* the space in s2 is irrelevant *)
        valid_spc t2 (pos2+1) t1 pos1 p
    else  (* the space in s2 is relevant, but no space in s1 !! *)
      check false
  else
    check (cc1 = cc2 && t1.tprop == t2.tprop) &&
    valid t1 (pos1+1) t2 (pos2+1)
    
        
      
let validate t1 t2 =
  let t1  = flatten_tdoc t1 in
  let t2 = flatten_tdoc t2 in
  (*
  Printf.printf "Compact"; print_newline ();
  Gc.compact ();
Printf.printf "done";  print_newline ();
  *)
  if t1.len > 0 && t2.len > 0 then 
    let s1 = t1.strs.(0) in
    let s2 = t2.strs.(0) in
    succ t1;
    succ t2;
    valid 
      { tstr = s1; tprop = t1.ints.(0); 
      tlen = String.length s1; tnext = t1 } 0
      { tstr = s2; tprop = t2.ints.(0); 
      tlen = String.length s2; tnext = t2 } 0
  else 
    t1.len = t2.len
      
    
  