open B8_types
open B8_types_zog


let dump_transition ?(chanout=stdout) t =
  List.iter 
    (fun trans ->
      output_string chanout
	("<"^(
	      match trans with
		SBold -> "B"
	      | EBold -> "/B"
	      | SEmph -> "EM"
	      | EEmph -> "/EM"
	      | SStrong -> "S"
	      | EStrong -> "/S"
	      | SItalic -> "I"
	      | EItalic -> "/I"
	      | SSize size -> String.make 1 (B8_dump.char_of_size size)
	      | ESize size -> "/"^(String.make 1 (B8_dump.char_of_size size))
	      | SColor c -> String.make 1 (B8_dump.char_of_color c)
	      | EColor c  -> "/"^(String.make 1 (B8_dump.char_of_color c))
	      | STT -> "TT"
	      | ETT -> "/TT"
	      | SU -> "U"
	      | EU -> "/U"
	      | SPL -> "PL"
	      | EPL -> "/PL"
	     )^
	 ">"
	)
    )
    t

let dump_doc ?(chanout=stdout) doc =
  List.iter
    (fun ele ->
      match ele with
	Text s -> output_string chanout s 
      |	Transition t -> dump_transition ~chanout: chanout t)
    doc

let head_true l =
  match l with
    b :: _ -> b
  | _ -> false

let head_int l =
  match l with
    h :: _ -> h
  | _ -> 0

let head_size l =
  match l with
    h :: _ -> h
  | _ -> Tnone

let head_color l =
  match l with
    c :: _ -> c
  | _ -> Unspecified

let tail_safe l =
  match l with
    _ :: q -> q
  | _ -> []

let pop_bool l =
  match l with 
    [] -> (false, [])
  | b :: q -> (b, q)

let pop_int l =
  match l with 
    [] -> (0, [])
  | n :: q -> (n, q)

let pop_trans (stack : trans_ele Stack.t) =
  try Stack.pop stack
  with Stack.Empty -> assert false

let top_trans (stack : trans_ele Stack.t) =
  try Stack.top stack
  with Stack.Empty -> assert false

let max x y = if x > y then y else x

let rec n_list n ele = 
  if n = 0 then [] else ele :: (n_list (n-1) ele)

let rec get_new_trans ctx trans_ele = 
  match trans_ele with

    SBold -> 
      let n = head_int ctx.b in
      let new_ctx = { ctx with 
		      b = (n+1) :: ctx.b ;
		      tag_stack = if n > 0 then ctx.tag_stack else SBold :: ctx.tag_stack 
		    } 
      in
      (new_ctx, if n > 0 then [] else [ SBold ])

  | EBold -> 
      let new_ctx = { ctx with b = tail_safe ctx.b } in
      (new_ctx, if head_int new_ctx.b > 0 then [] else [ EBold ])

  | SEmph ->
      if head_int ctx.s > 0 then
	(ctx, [])
      else
	(
	 let b = head_true ctx.em in 
	 let new_ctx = { ctx with
			 em = (not b) :: ctx.em ;
			 tag_stack = SEmph :: ctx.tag_stack 
		       } 
	 in
	 (new_ctx, [ SEmph ]) (* A AMELIORER *)
	)
  | EEmph -> 
      if head_int ctx.s > 0 then
	(ctx, [])
      else
	(
	 let new_ctx = { ctx with em = tail_safe ctx.em } in
	 (new_ctx, [ EEmph ])
	)

  | SStrong -> 
      let n = head_int ctx.s in
      let new_ctx = { ctx with 
		      s = (n+1) :: ctx.s ;
		      tag_stack = if n > 0 then ctx.tag_stack else SStrong :: ctx.tag_stack 
		    } 
      in
      (new_ctx, if n > 0 then [] else [ SStrong ])

  | EStrong -> 
      let new_ctx = { ctx with s = tail_safe ctx.s } in
      (new_ctx, if head_int new_ctx.s > 0 then [] else [ EStrong ])

  | SItalic -> 
      let n = head_int ctx.i in
      let new_ctx = { ctx with 
		      i = (n+1) :: ctx.i ;
		      tag_stack = if n > 0 then ctx.tag_stack else SItalic :: ctx.tag_stack 
		    } 
      in
      (new_ctx, if n > 0 then [] else [ SItalic ])

  | EItalic ->
      let new_ctx = { ctx with i = tail_safe ctx.i } in
      (new_ctx, if head_int new_ctx.i > 0 then [] else [ EItalic ])

  | SSize size -> 
      (
       match ctx.size with
	 [] -> 
	   let new_ctx = { ctx with 
			   size = [ size ] ;
			   tag_stack = trans_ele :: ctx.tag_stack 
			 } 
	   in
	   (new_ctx, [ trans_ele ])
       | s :: q ->
	   let new_ctx = { ctx with 
			   size = size :: s :: q ;
			   tag_stack = if size = s then ctx.tag_stack else trans_ele :: ctx.tag_stack
			 } 
	   in
	   if size = s then 
	     (new_ctx, [])
	   else
	     (new_ctx, [ trans_ele ])
      )
  | ESize size ->
      (
       match ctx.size with
	 _ :: q ->
	 (* we don't look but the popped size should be one we have ni ESize *)
	   let new_ctx = { ctx with size = q } in
	   (
	    match q with
	      s :: _ ->
		(new_ctx, if size = s then [] else [ trans_ele ])
	    | _ ->
		(new_ctx, [ trans_ele ])
	   )
       |  _ ->
	   (ctx, [])
      )

  | SColor color ->
      (
       match ctx.color with
	 [] -> 
	   let new_ctx = { ctx with 
			   color = [ color ] ;
			   tag_stack = trans_ele :: ctx.tag_stack 
			 }
	   in
	   (new_ctx, [ trans_ele ])
       | c :: q ->
	   let new_ctx = { ctx with 
			   color = color :: c :: q ;
			   tag_stack = if color = c then ctx.tag_stack else trans_ele :: ctx.tag_stack 
			 } 
	   in
	   if color = c then 
	     (new_ctx, [])
	   else
	     (new_ctx, [ trans_ele ])
      )
  | EColor color ->
      (
       match ctx.color with
	 _ :: q ->
	 (* we don't look but the popped size should be one we have ni EColor *)
	   let new_ctx = { ctx with color = q } in
	   (
	    match q with
	      c :: _ ->
		(new_ctx, if color = c then [] else [ trans_ele ])
	    | _ ->
		(new_ctx, [ trans_ele ])
	   )
       |  _ ->
	   (ctx, [])
      )

  | STT -> 
      let n = head_int ctx.tt in
      let new_ctx = { ctx with 
		      tt = (n+1) :: ctx.tt ;
		      tag_stack = if n > 0 then ctx.tag_stack else STT :: ctx.tag_stack 
		    } 
      in
      (new_ctx, if n > 0 then [] else [ STT ])

  | ETT ->
      let new_ctx = { ctx with tt = tail_safe ctx.tt } in
      (new_ctx, if head_int new_ctx.tt > 0 then [] else [ ETT ])

  | SU -> 
      let n_previous = head_int ctx.u in
      let new_ctx = { ctx with 
		      u = (n_previous+1) :: ctx.u ; 
		      tag_stack = if (n_previous +1) > 3 then ctx.tag_stack else trans_ele :: ctx.tag_stack
		    }
      in
      (* si on dpase 3, on ne garde pas la transition, et
	 comme on a + de trois dans la pile, on saura qu'il faut aussi
	 virer le prochain tag fermant. *)
      (new_ctx, if (n_previous + 1) > 3 then [] else [ trans_ele ])
      
  | EU ->
       let (new_ctx, n) = 
	 match ctx.u with
	   n :: q ->
	     ({ ctx with u = q }, n)
	 | _ ->
	     (ctx, 0) (* bizarre si on est l *)
       in
       (new_ctx, if n > 3 then [] else [ trans_ele ])

  | SPL ->
      let b_previous = head_int ctx.b in
      let i_previous = head_int ctx.i in
      let em_previous = head_true ctx.em in
      let s_previous = head_int ctx.s in
      let u_previous = head_int ctx.u in
      let tt_previous = head_int ctx.tt in
      let new_ctx = { ctx with 
		      b = 0 :: ctx.b ;
		      i = 0 :: ctx.i ;
		      em = false :: ctx.em ;
		      s = 0 :: ctx.s ;
		      u = 0 :: ctx.u ;
		      tt = 0 :: ctx.tt ;
		      tag_stack = SPL :: ctx.tag_stack 
		    }
      in
      (* on clate PL pour mettre un tag pour les trucs qui ont chang *)
      (* il faut fermer n fois les bold, avec n le nombre de bold ouverts *)
      (new_ctx, [ SPL ])
(*       (if b_previous > 0 then [ EBold ] else [] ) @
       (if i_previous > 0 then [ EItalic ] else [] ) @
       (if s_previous > 0 then [ EStrong ] else [] ) @
       (if tt_previous > 0 then [ ETT ] else [] ) @
       (n_list (if u_previous > 3 then 3 else u_previous) EU ) @
       (if em_previous then [ EEmph ] else [])
      )
*)

  | EPL ->
      let (b_pop, b_q) = pop_int ctx.b in
      let b_previous = head_int b_q in
      let (i_pop, i_q) = pop_int ctx.i in
      let i_previous = head_int i_q in
      let (em_pop, em_q) = pop_bool ctx.em in
      let em_previous = head_true em_q in
      let (s_pop, s_q) = pop_int ctx.s in
      let s_previous = head_int s_q in
      let (u_pop, u_q) = pop_int ctx.u in
      let u_previous = head_int u_q in
      let (tt_pop, tt_q) = pop_int ctx.tt in
      let tt_previous = head_int tt_q in
      let new_ctx = { ctx with
		      b = b_q ;
		      i = i_q ;
		      em = em_q ;
		      tt = tt_q ;
		      s = s_q ;
		      u = u_q }
      in
      (new_ctx, [ EPL ])
(*       (if b_previous > 0 then [ SBold ] else [] ) @
       (if i_previous > 0 then [ SItalic ] else [] ) @
       (if s_previous > 0 then [ SStrong ] else [] ) @
       (if tt_previous > 0 then [ STT ] else [] ) @
       (n_list (if u_previous > 3 then 3 else u_previous) SU ) @
       (if em_previous then [ SEmph ] else [])
      )
*)

  | _ ->
      (ctx, [trans_ele])


let not_blank = Str.regexp "[^ \013\009\012]+"
let blanks = Str.regexp "[ \013\009\012]+"
let begin_blanks = Str.regexp "^[ \013\009\012]+"
let end_blanks = Str.regexp "[ \013\009\012]+$"

let remove_double_blanks s = Str.global_replace blanks " " s

let rec remove_first l ele =
  match l with
    [] -> (false, [])
  | e :: q when e = ele -> (true, q)
  | e :: q -> 
      let (b, q2) = remove_first q ele in
      (b, e :: q2)

let inverse_trans_ele trans_ele =
  match trans_ele with
    SBold -> EBold
  | EBold -> SBold
  | SEmph -> EEmph
  | EEmph -> SEmph
  | SStrong -> EStrong
  | EStrong -> SStrong
  | SItalic -> EItalic
  | EItalic -> SItalic
  | SSize size -> ESize size
  | ESize size -> SSize size
  | SColor color -> EColor color
  | EColor color  -> SColor color
  | STT -> ETT
  | ETT -> STT
  | SU -> EU
  | EU -> SU
  | SPL -> EPL
  | EPL -> SPL


(* Transfert des blancs en tte de s2 vers s1. *)
let transfert_blancs s1 s2 =
  prerr_endline ("transfert_blancs s1="^s1^", s2="^s2);
  (s1, s2)

(* Optimizing in a transition : <I>...</I>, or </I>...<I> are removed.
   Warning : There must not be any PL in the transition. *)
let simplify_transition t =
  let rec iter acc = function
      [] -> acc
    | ele :: q -> 
	let inverse_ele = inverse_trans_ele ele in
	let (changed, q2) = remove_first q inverse_ele in
	iter (if changed then acc else acc @ [ele]) q2 
  in
  iter [] t

let rec optimize ctx doc =
  match doc with
  | [] ->
(*      prerr_endline "F0"; dump_doc ~chanout: stderr doc ;*)
      []
  | (Text s1) :: (Text s2) :: q ->
(*      prerr_endline "F1"; dump_doc ~chanout: stderr doc ;*)
      optimize ctx ((Text (s1^s2)) :: q)

  | (Transition t1) :: (Transition t2) :: q ->
(*      prerr_endline "F2"; dump_doc ~chanout: stderr doc ;*)
      optimize ctx (Transition (t1 @ t2) :: q)

  | (Transition t1) :: (((Text "") :: (Transition t2) :: (Text s2) :: q) as q2) ->
(*      prerr_endline "F3"; dump_doc ~chanout: stderr doc ;*)
      optimize ctx ((Transition (t1 @ t2)) :: (Text s2) :: q)

  | (Transition t1) :: (((Text s1) :: (Transition t2) :: (Text s2) :: q) as q2) ->
(*      prerr_endline "F4"; dump_doc ~chanout: stderr doc ;*)
      let (ctx1, new_t1) = optimize_transition ctx t1 in
      let (ctx2, new_t2) = optimize_transition ctx1 t2 in
      if new_t2 = [] then
	(
	 (Transition new_t1) :: (optimize ctx2 ((Text (s1^s2)) :: q))
	)
      else
	(
	 (* on peut faire le merge si le tt et la taille ne change pas entre t1 et t2 *)
	 let n_tt1 = head_int ctx1.tt in
	 let n_tt2 = head_int ctx2.tt in
	 let size1 = head_size ctx1.size in
	 let size2 = head_size ctx2.size in
	 let u1 = max (head_int ctx1.u) 3 in
	 let u2 = max (head_int ctx2.u) 3 in
	 let c1 = head_color ctx1.color in
	 let c2 = head_color ctx2.color in
	 if ((n_tt1 = 0 & n_tt2 = 0) or 
	     (n_tt1 > 0) & (n_tt2 > 0)) &
	   ( size1 = size2) &
	   ( u1 = u2 ) &
	   (( u1 = 0) or ( c1 = c2 ))
	 then
	   try
	     let _ = Str.search_forward not_blank s2 0 in
	     try
	       (* on essaie maintenant de coller s1  s2 *)
	       let _ = Str.search_forward not_blank s1 0 in
	       (* on a trouv un caractre non blanc *)
	       (* on peut quand mme 'pomper' les blancs de s2 pour les mettre dans s1 *)
	       let (new_s1, new_s2) = s1, s2 in (*transfert_blancs s1 s2 in*)
	       (Transition new_t1) :: 
	       (optimize ctx1 ((Text new_s1) :: (Transition t2) :: (Text new_s2) :: q))
	     with
	       Not_found ->
		 optimize ctx ((Transition (t1 @ t2)) :: (Text (s1^s2)) :: q)
	   with
	     Not_found ->
	       (* on peut coller s2  la suite de s1 *)
	       prerr_endline "F4_4";
	       prerr_endline ("s2="^s2);
	       let new_s1 = s1^s2 in
	       (Transition new_t1) :: 
	       (optimize ctx1 ((Text new_s1) :: (Transition t2) :: q))
	       
	 else
	   (Transition new_t1) :: (optimize ctx1 q2)
	)

  | (Transition t) :: q ->
(*      prerr_endline "F5"; dump_doc ~chanout: stderr doc ;*)
      let (new_ctx, new_tr) = optimize_transition ctx t in
      (Transition new_tr) :: (optimize new_ctx q)

  | (Text s) :: q ->
(*      prerr_endline "F6"; dump_doc ~chanout: stderr doc ;*)
      let s2 = 
	if head_int ctx.tt > 0 then
	  s
	else
	  remove_double_blanks s
      in
      (Text s2) :: (optimize ctx q)

and optimize_transition ctx t =
  let (new_ctx, new_transition) = 
    List.fold_left 
      (fun (acc_ctx, acc) -> fun t_ele ->
	let (new_ctx, new_t) = get_new_trans acc_ctx t_ele in
	(new_ctx, acc @ new_t)
      )
      (ctx, [])
      t
  in
(*  let new_transition2 = simplify_transition new_transition in*)
  (new_ctx, new_transition)

(** EM Elimination *)
(*
(* Retourne true et la liste des trans_ele restant avant et aprs le premier Semph trouv,
   ou ([], false, []). *)
let rec get_first_Semph t =
  match t with
    [] -> ([], false, [])
  | SEmph :: q ->
      (acc, true, q)
  | ele :: q ->
      get_first_Semph (acc @ [ele]) q

let rec remove_next_Semph_in_transition acc n t =
  match t with
    [] ->
      (acc, false, t)
  | SEmph :: q ->
      if n > 0 then
	remove_next_Semph_in_transition (acc @ [SEmph]) (n+1) q
      else
	(acc, true, q)
  | EEmph :: q ->
      if n > 0 then
	remove_next_Semph_in_transition (acc @ [EEmph]) (n-1) q
      else
	(* on a perdu : le Emph est ferm avant qu'on en ait trouv un autre *)
	(acc, false, t)
  | (SPL as ele) :: q
  | (SStrong as ele) :: q ->
      remove_next_Semph_in_transition (acc @ [ele]) (n+1) q
  | (EPL as ele) :: q
  | (EStrong as ele) :: q ->
      remove_next_Semph_in_transition (acc @ [ele]) (n-1) q
  | ele :: q ->
      remove_next_Semph_in_transition (acc@ [ele]) n q

(* Enlve le prochain SEmph annulant celui duquel on est parti,
   si on le trouve. Retourne (true/false, document). n est le
   nombre de EEmph, EPL et EStrong  laisser passer ; il est incrment quand on
   rencontre un SEmph alors que n > 0, un SStrong, ou un SPL. *)
let rec remove_next_Semph acc n doc =
  match doc with
    [] -> acc
  | (Text s as ele) :: q ->
      (
       (* s'il n'y a pas que des blancs, on s'arrte. *)
       let n_opt = 
	 try Some (Str.search_forward not_blank s2 0)
	 with Not_found -> None
       in
       match n_opt with
	 None ->
	   (* on n'a que des blancs, on peut continuer *)
	   remove_next_Semph (acc @ [ele]) n q
       | Some _ ->
	   (* on s'arrte l *)
	   (false, doc)
      )
  | (Transition t as trans) :: q ->
      let (t2_avant, b_remove, t2_apres) = remove_next_Semph_in_transition [] 0 t in
      if b_remove then
	let (remain_avant, b_remove_in_trans1, remain_apres) = 
	  remove_next_Eemph_in_transition [] 0 t2_apres 
	in
	if b_remove_in_trans1 then
	  (* j'en ai vir un, on va virer le second *)
	  let (remain_avant2, b_remove_in_trans2, remain_apres2) = remove_next_Eemph_in_transition [] 0 remain_apres in
	  if b_remove_in_trans2 then
	    (* on a vir les deux lments *)
	    (Transition (t2_avant @ remain_avant @ remain_avant2 @ remain_apres2)) :: q
	  else
	    (* on cherche encore un EEmph *)
	    (Transition (t2_avant @ remain_avant @ remain_apres)) ::
	    (remove_next_Eemph [] 1 0 q)
	else
	  (* on cherche 2 EEmph *)
	  trans :: (remove_next_Eemph [] 1 0 q)
      else
        (* on cherche 2 EEmph *)
	remove_next_Semph (acc @ [trans]) 
      
(* EM elimination *)
let rec em_elimination ctx doc =
  match doc with
    [] ->
      []
  | (Text _ as ele) :: q ->
      ele :: (em_elimination ctx doc)
  | (Transition t as ele) :: q ->
      let (avant, b, apres) = get_first_Semph [] t in
      if b then
	(
	 (* aller supprimer le EMph suivant et les /EM, sauf si
	    arrt  cause de texte... *)
	 let (t2_avant, b_remove, t2_apres) = remove_next_Semph_in_transition [] 0 apres in
	 if b_remove then
	   let (remain_avant, b_remove_in_trans1, remain_apres) = 
	     remove_next_Eemph_in_transition [] 0 t2_apres 
	   in
	   if b_remove_in_trans1 then
	     (* j'en ai vir un, on va virer le second *)
	     let (remain_avant2, b_remove_in_trans2, remain_apres2) = remove_next_Eemph_in_transition [] 0 remain_apres in
	     if b_remove_in_trans2 then
	       (* on a vir les deux lments *)
	       (Transition (avant @ t2_avant @ remain_avant @ remain_avant2 @ remain_apres2)) :: q
	     else
	       (* on cherche encore un EEmph *)
	       (Transition (avant @ t2_avant @ remain_avant @ remain_apres)) ::
	       (remove_next_Eemph [] 1 0 q)
	   else
	     (* on cherche 2 EEmph *)
	     (Transition avant @ apres) :: (remove_next_Eemph [] 2 0 q)
	 else
           (* on cherche 2 EEmph *)
	   (Transition (avant@apres)) @ (remove_next_Eemph [] 2 0 q)
	)
      else
	ele :: (em_elimination ctx q)

*)
let lexbuf = Lexing.from_channel stdin 
let doc = B8_parser_zog.main  B8_lexer_zog.token lexbuf 

(*let _ = dump_doc doc*)
let doc2 = optimize (B8_types_zog.new_context ()) doc 
let _ = dump_doc doc2

(*let doc3 = optimize (B8_types_zog.new_context ()) doc2
let _ = dump_doc doc3
*)
